use std::time::Instant;

use anyhow::{bail, Context, Error};

pub mod templates;

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum LispVMData {
    Sample(u32),
    // TODO: MIDI Packet, lighting packet, etc
}

impl LispVMData {
    pub fn try_get_sample(self) -> Result<u32, Error> {
        match self {
            LispVMData::Sample(x) => Ok(x),
            _ => bail!(
                "Lisp VM attempted to convert {:?} to Sample data, it is not",
                self
            ),
        }
    }
}

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum OpCode {
    Add,
    AddConst(u32),
    Sub,
    SubConst(u32),
    Mul,
    Div,
    Jump(usize),
    JumpIfZero(usize),
    JumpIfEq(usize, LispVMData),
    JumpIfStackSizeEq(usize, usize),
    Input(LispVMData),
    Output(LispVMData),
    Pop,
    Dup,
    Brk,
    MicrosElapsedModN(u32),
}

impl OpCode {
    pub fn try_get_data(self) -> Result<LispVMData, Error> {
        match self {
            OpCode::Input(x) => Ok(x),
            OpCode::Output(x) => Ok(x),
            _ => bail!("Failed to coerce opcode {:?} to data", self),
        }
    }
    pub fn try_get_sample(self) -> Result<u32, Error> {
        self.try_get_data()
            .context(format!("Failed to coerce opcode {:?} as sample data", self))?
            .try_get_sample()
    }
}

pub struct Script {
    bytecode: Vec<OpCode>,
    source: Option<String>,
}

impl Script {
    pub fn from_lisp(text: String) -> Self {
        let opcodes: Vec<OpCode> = vec![];

        // TODO: actual parse that shit

        Self {
            bytecode: opcodes,
            source: Some(text),
        }
    }

    pub fn from_raw_bytecode(bytecode: Vec<OpCode>) -> Self {
        Self {
            bytecode,
            source: None,
        }
    }

    pub fn process(self, input: Vec<OpCode>, time: Instant) -> Result<Vec<OpCode>, Error> {
        let mut stack: Vec<OpCode> = input;
        let mut pc = 0;
        let mut dont_increment_pc = false;
        while pc < self.bytecode.len() {
            match self.bytecode.get(pc).unwrap() {
                OpCode::Add => {
                    let a = stack
                        .pop()
                        .context(format!("Stack underflow on add at {}", pc))?
                        .try_get_sample()?;
                    let b = stack
                        .pop()
                        .context(format!("Stack underflow on add at {}", pc))?
                        .try_get_sample()?;
                    stack.push(OpCode::Output(LispVMData::Sample(a + b)));
                }
                OpCode::AddConst(x) => {
                    let a = stack
                        .pop()
                        .context(format!("Stack underflow on add at {}", pc))?
                        .try_get_sample()?;
                    stack.push(OpCode::Output(LispVMData::Sample(a + x)));
                }
                OpCode::Sub => {
                    let b = stack
                        .pop()
                        .context(format!("Stack underflow on sub at {}", pc))?
                        .try_get_sample()?;
                    let a = stack
                        .pop()
                        .context(format!("Stack underflow on sub at {}", pc))?
                        .try_get_sample()?;
                    stack.push(OpCode::Output(LispVMData::Sample(a - b)));
                }
                OpCode::SubConst(x) => {
                    let a = stack
                        .pop()
                        .context(format!("Stack underflow on sub at {}", pc))?
                        .try_get_sample()?;
                    stack.push(OpCode::Output(LispVMData::Sample(a - x)));
                }
                OpCode::Mul => todo!(),
                OpCode::Div => todo!(),
                OpCode::Jump(x) => {
                    if *x >= self.bytecode.len() {
                        bail!(
                            "Jump to {} is out of bounds. Code is {} opcodes long",
                            *x,
                            self.bytecode.len()
                        );
                    } else {
                        pc = *x;
                        dont_increment_pc = true;
                    }
                }
                OpCode::JumpIfZero(x) => {
                    if stack
                        .last()
                        .context(format!("Stack underflow on jz at {}", pc))?
                        .try_get_sample()?
                        == 0
                    {
                        if *x >= self.bytecode.len() {
                            bail!(
                                "Jump to {} is out of bounds. Code is {} opcodes long",
                                *x,
                                self.bytecode.len()
                            );
                        } else {
                            pc = *x;
                            dont_increment_pc = true;
                        }
                    }
                }
                OpCode::JumpIfEq(x, y) => {
                    if stack
                        .last()
                        .context(format!("Stack underflow on jq at {}", pc))?
                        .try_get_data()?
                        == *y
                    {
                        if *x >= self.bytecode.len() {
                            bail!(
                                "Jump to {} is out of bounds. Code is {} bytes long",
                                *x,
                                self.bytecode.len()
                            );
                        } else {
                            pc = *x;
                            dont_increment_pc = true;
                        }
                    }
                }
                OpCode::JumpIfStackSizeEq(x, y) => {
                    if stack.len() == *y {
                        if *x >= self.bytecode.len() {
                            bail!(
                                "Jump to {} is out of bounds. Code is {} bytes long",
                                *x,
                                self.bytecode.len()
                            );
                        } else {
                            pc = *x;
                            dont_increment_pc = true;
                        }
                    }
                }
                OpCode::Input(x) => {
                    stack.push(OpCode::Output(*x));
                }
                OpCode::Output(x) => {
                    stack.push(OpCode::Output(*x));
                }
                OpCode::Pop => {
                    stack.pop();
                }
                OpCode::Dup => {
                    stack.push(
                        *stack
                            .last()
                            .context(format!("Stack underflow on add at dup {}", pc))?,
                    );
                }
                OpCode::Brk => {
                    return Ok(stack);
                }
                OpCode::MicrosElapsedModN(x) => {
                    stack.push(OpCode::Output(LispVMData::Sample(u32::try_from(time.elapsed().as_micros() % (*x as u128)).unwrap())));
                },
            }
            if dont_increment_pc {
                dont_increment_pc = false;
            } else {
                pc += 1;
            }
        }
        Ok(stack)
    }
}
